@@ -225,7 +225,7 @@ __all__ = ["compile", "escape", "findall", "finditer", "fullmatch", "match",
   "V0", "VERSION0", "V1", "VERSION1", "X", "VERBOSE", "W", "WORD", "error",
   "Regex"]
 
-__version__ = "2.4.18"
+__version__ = "2.4.38"
 
 # --------------------------------------------------------------------
 # Public interface.
@@ -344,7 +344,7 @@ def escape(pattern, special_only=False):
         s = []
         if special_only:
             for c in pattern:
-                if c in _NONLITERAL:
+                if c in _METACHARS:
                     s.append("\\")
                     s.append(c)
                 elif c == "\x00":
@@ -366,7 +366,7 @@ def escape(pattern, special_only=False):
         s = []
         if special_only:
             for c in pattern:
-                if chr(c) in _NONLITERAL:
+                if chr(c) in _METACHARS:
                     s.extend(b"\\")
                     s.append(c)
                 elif c == 0:
@@ -390,6 +390,7 @@ def escape(pattern, special_only=False):
 
 import _regex_core
 import _regex
+from threading import RLock as _RLock
 from _regex_core import *
 from _regex_core import (_ALL_VERSIONS, _ALL_ENCODINGS, _FirstSetError,
   _UnscopedFlagSet, _check_group_features, _compile_firstset,
@@ -403,82 +404,83 @@ from _regex_core import (ALNUM as _ALNUM, Info as _Info, OP as _OP, Source as
 
 DEFAULT_VERSION = VERSION0
 
-_NONLITERAL = frozenset("()[]{}?*+|^$\\.")
+_METACHARS = frozenset("()[]{}?*+|^$\\.")
 
 _regex_core.DEFAULT_VERSION = DEFAULT_VERSION
 
 # Caches for the patterns and replacements.
 _cache = {}
+_cache_lock = _RLock()
 _named_args = {}
+_replacement_cache = {}
 
 # Maximum size of the cache.
 _MAXCACHE = 500
+_MAXREPCACHE = 500
 
-def _compile(pattern, flags=0, kwargs=None):
+def _compile(pattern, flags=0, kwargs={}):
     "Compiles a regular expression to a PatternObject."
-    # We're checking in this order because _pattern_type isn't defined when
-    # _compile() is first called, with a string pattern, but only after the
-    # support objects are defined.
-    if isinstance(pattern, (str, bytes)):
+    try:
+        # Do we know what keyword arguments are needed?
+        args_key = pattern, type(pattern), flags
+        args_needed = _named_args[args_key]
+
+        # Are we being provided with its required keyword arguments?
+        args_supplied = set()
+        if args_needed:
+            for k, v in args_needed:
+                try:
+                    args_supplied.add((k, frozenset(kwargs[k])))
+                except KeyError:
+                    raise error("missing named list")
+
+        args_supplied = frozenset(args_supplied)
+
+        # Have we already seen this regular expression and named list?
+        pattern_key = (pattern, type(pattern), flags, args_supplied,
+          DEFAULT_VERSION)
+        return _cache[pattern_key]
+    except KeyError:
+        # It's a new pattern, or new named list for a known pattern.
         pass
+
+    # Guess the encoding from the class of the pattern string.
+    if isinstance(pattern, str):
+        guess_encoding = UNICODE
+    elif isinstance(pattern, bytes):
+        guess_encoding = ASCII
     elif isinstance(pattern, _pattern_type):
         if flags:
             raise ValueError("can't process flags argument with a compiled pattern")
+
         return pattern
     else:
         raise TypeError("first argument must be a string or compiled pattern")
 
-    if kwargs is None:
-        kwargs = {}
-
     # Set the default version in the core code in case it has been changed.
     _regex_core.DEFAULT_VERSION = DEFAULT_VERSION
 
-    # Do we know what keyword arguments are needed?
-    args_key = pattern, type(pattern), flags
-    args_needed = _named_args.get(args_key)
-    if args_needed is not None:
-        # Are we being provided with those keyword arguments?
-        args_supplied = set()
-        for k, v in args_needed:
-            if k not in kwargs:
-                raise error("missing named list")
-            args_supplied.add((k, frozenset(kwargs[k])))
-
-        # Have we already seen this regular expression and named list?
-        pattern_key = (pattern, type(pattern), flags, frozenset(args_supplied),
-          DEFAULT_VERSION)
-        compiled_pattern = _cache.get(pattern_key)
-        if compiled_pattern:
-            return compiled_pattern
-
-    # It's a new pattern (or new named list for a known pattern).
-
-    # Guess the encoding from the class of the pattern string.
-    if isinstance(pattern, bytes):
-        guess_encoding = ASCII
-    else:
-        guess_encoding = UNICODE
-
-    # Parse the pattern. In the old behaviour the inline flags are global and
-    # the pattern will need to be reparsed if a flag becomes turned on.
-    version = (flags & _ALL_VERSIONS) or DEFAULT_VERSION
-    global_flags = flags | _regex_core.DEFAULT_FLAGS.get(version, 0)
+    caught_exception = None
 
     while True:
         try:
             source = _Source(pattern)
-            info = _Info(global_flags, source.char_type, kwargs)
+            info = _Info(flags, source.char_type, kwargs)
             info.guess_encoding = guess_encoding
             source.ignore_space = bool(info.flags & VERBOSE)
             parsed = _parse_pattern(source, info)
             break
-        except _UnscopedFlagSet as e:
+        except _UnscopedFlagSet:
             # Remember the global flags for the next attempt.
-            global_flags = e.global_flags | flags
+            flags = info.global_flags
+        except error as e:
+            caught_exception = e
+
+        if caught_exception:
+            raise error(str(caught_exception))
 
     if not source.at_end():
-        raise error("trailing characters in pattern")
+        raise error("trailing characters in pattern at position {}".format(source.pos))
 
     # Check the global flags for conflicts.
     version = (info.flags & _ALL_VERSIONS) or DEFAULT_VERSION
@@ -500,6 +502,10 @@ def _compile(pattern, flags=0, kwargs=None):
     reverse = bool(info.flags & REVERSE)
     fuzzy = isinstance(parsed, _Fuzzy)
 
+    # Should we print the parsed pattern?
+    if flags & DEBUG:
+        parsed.dump(indent=0, reverse=reverse)
+
     # Fix the group references.
     parsed.fix_groups(reverse, False)
 
@@ -525,10 +531,6 @@ def _compile(pattern, flags=0, kwargs=None):
         named_list_indexes[index] = items
         args_needed.add((name, values))
 
-    # Should we print the parsed pattern?
-    if flags & DEBUG:
-        parsed.dump(indent=0, reverse=reverse)
-
     # Check the features of the groups.
     _check_group_features(info, parsed)
 
@@ -570,11 +572,12 @@ def _compile(pattern, flags=0, kwargs=None):
     # affect the code generation but _are_ needed by the PatternObject.
     compiled_pattern = _regex.compile(pattern, info.flags | version, code,
       info.group_index, index_group, named_lists, named_list_indexes,
-      req_offset, req_chars, req_flags)
+      req_offset, req_chars, req_flags, info.group_count)
 
     # Do we need to reduce the size of the cache?
     if len(_cache) >= _MAXCACHE:
-        _shrink_cache(_cache, _named_args, _MAXCACHE)
+        with _cache_lock:
+            _shrink_cache(_cache, _named_args, _MAXCACHE)
 
     args_needed = frozenset(args_needed)
 
@@ -590,6 +593,16 @@ def _compile(pattern, flags=0, kwargs=None):
 def _compile_replacement_helper(pattern, template):
     "Compiles a replacement template."
     # This function is called by the _regex module.
+
+    # Have we seen this before?
+    key = pattern.pattern, pattern.flags, template
+    compiled = _replacement_cache.get(key)
+    if compiled is not None:
+        return compiled
+
+    if len(_replacement_cache) >= _MAXREPCACHE:
+        _replacement_cache.clear()
+
     is_unicode = isinstance(template, str)
     source = _Source(template)
     if is_unicode:
@@ -624,6 +637,9 @@ def _compile_replacement_helper(pattern, template):
     # Flush the literal.
     if literal:
         compiled.append(make_string(literal))
+
+    _replacement_cache[key] = compiled
+
     return compiled
 
 # We define _pattern_type here after all the support objects have been defined.
